from __future__ import annotations
from pathlib import Path

from pybi.core import DataSource
import pandas as pd

from typing import TYPE_CHECKING, Dict, List, Optional, Union
import os

from pybi.core.components import (
    ContainerComponent,
    ComponentTag,
    BoxComponent,
    FlowBoxComponent,
    TextComponent,
    UploadComponent,
    GridBoxComponent,
    TabsComponent,
    Markdown,
    IconComponent,
    SvgIconComponent,
    AffixComponent,
    Mermaid,
    Input,
    NumberSlider,
    SpaceComponent,
)
from pybi.core.components.reactiveComponent import EChart, Slicer, Table, TextValue
from pybi.core.dataSource import (
    DataSourceField,
    DataSourceTable,
    DataView,
    DataViewBase,
    PivotDataView,
)
from pybi.utils.dataSourceUtils import ds2sqlite_file_base64, ds2sqlite
from pybi.utils.data_gen import (
    JsonUtils,
    random_ds_name,
    random_dv_name,
    get_project_root,
)
from pybi.utils.markdown2 import markdown
from pybi.core.uiResource import ResourceManager

from pybi.core.sql import SqlInfo, SqlWrapper
import pybi.utils.sql as sqlUtils
from pybi.easyEcharts.base import BaseChart, ChartCollector
from pybi.easyEcharts import easy_echarts
from pybi.core.components.reactiveComponent.echarts import (
    EChartInfo,
    EChartDatasetInfo,
    EChartJscode,
    OptionsExtractor,
)
import pybi.utils.pyecharts_utils as pyecharts_utils
from pybi.icons.iconManager import get_singleton as get_iconManager
from pybi.core.webResources import WebResourceManager
from pybi.core.styles.styleTag import StyleTagInfo
import pybi as pbi

if TYPE_CHECKING:
    pass
    # from pybi.core.components import ReactiveComponent


class AppMeta:
    def __init__(self, app: App) -> None:
        self.__app = app

    def set_dbLocalStorage(self, on: bool):
        """
        是否开启数据库本地缓存
        """
        self.__app.dbLocalStorage = on
        return self

    def set_echarts_renderer(self, renderer="canvas"):
        """
        echarts renderer type: 'canvas' or 'svg'
        'canvas' is default
        """
        self.__app.echartsRenderer = renderer
        return self

    def set_doc_title(self, title: str):
        self.__app._doc_title = title
        return self

    def add_style_tag(
        self, css: str, id: Optional[str] = None, media: Optional[str] = None
    ):
        self.__app._style_tags.append(StyleTagInfo(css, id, media))
        return self


class App(ContainerComponent):
    def __init__(self) -> None:
        super().__init__(ComponentTag.App)
        self.dataSources: List[DataSource] = []
        self.dataViews: List[DataViewBase] = []
        self.__dataSetNames = set()
        self._with_temp_host_stack: List[ContainerComponent] = []
        self._clear_data = False
        self.dbLocalStorage = False
        self.echartsRenderer = "canvas"
        self._doc_title: Optional[str] = None
        self.__meta = AppMeta(self)
        self.__json_utils = JsonUtils()
        self.__resourceManager = ResourceManager()
        self._iconManager = get_iconManager()

        self.__webResourceManager = WebResourceManager()
        self.webResources = []
        self._style_tags: List[StyleTagInfo] = []

    def __record_and_check_dataset_name(self, name: str):
        if name in self.__dataSetNames:
            raise Exception(f"dataset name '{name}' is duplicate")
        self.__dataSetNames.add(name)

    @property
    def meta(self):
        return self.__meta

    def clear_all_data(self):
        self._clear_data = True

    def _get_temp_host(self):
        if self._with_temp_host_stack:
            return self._with_temp_host_stack[len(self._with_temp_host_stack) - 1]
        return None

    def set_source(self, data: pd.DataFrame, *, name: Optional[str] = None):
        name = name or random_ds_name()
        self.__record_and_check_dataset_name(name)

        ds = DataSource(name, data)
        self.dataSources.append(ds)
        return DataSourceTable(ds.name, data.columns.tolist(), host=self)

    def set_dataView(
        self,
        sql: str,
        exclude_source: Optional[List[DataSourceTable]] = None,
        *,
        name: Optional[str] = None,
    ):
        exclude_source = exclude_source or []
        name = name or random_dv_name()
        self.__record_and_check_dataset_name(name)
        dv = DataView(name, sql)

        for es in exclude_source:
            dv.exclude_source(es.source_name)

        self.dataViews.append(dv)
        return DataSourceTable(
            dv.name, sqlUtils.extract_fields_head_select(sql), host=self
        )

    def set_pivot_dataView(
        self,
        source: str,
        row: str,
        column: str,
        cell: str,
        agg="min",
        exclude_source: Optional[List[DataSourceTable]] = None,
        excludeRowFields=False,
        *,
        name: Optional[str] = None,
    ):
        exclude_source = exclude_source or []
        name = name or random_dv_name()
        self.__record_and_check_dataset_name(name)
        pdv = PivotDataView(name, source, row, column, cell, agg, excludeRowFields)

        for es in exclude_source:
            pdv.exclude_source(es.source_name)

        self.dataViews.append(pdv)
        return DataSourceTable(pdv.name, [], host=self)

    def sql(self, sql: str):
        return SqlWrapper(sql)

    def add_upload(
        self,
        *,
        host: Optional[ContainerComponent] = None,
    ):
        cp = UploadComponent()

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        self.__resourceManager.register_element_cps()
        return cp

    def add_text(
        self,
        text: Union[str, SqlWrapper],
        *,
        host: Optional[ContainerComponent] = None,
    ):
        if isinstance(text, SqlWrapper):
            text = str(text)
        contexts = list(SqlInfo.extract_sql_from_text(text))

        cp = TextValue(contexts)

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp

    def add_slicer(
        self,
        field: Union[DataSourceField, DataSourceTable],
        *,
        orderby: Optional[str] = None,
        host: Optional[ContainerComponent] = None,
    ):
        """
        orderby:
            Defaults `None`:  select distinct colA from data
            '1' : select distinct colA from data order by 1
            '1 desc' : select distinct colA from data order by 1 desc
            'colX dsce' : select distinct colA from data order by colX desc
        """
        if isinstance(field, DataSourceTable):
            field = field[field.columns[0]]

        assert isinstance(field, DataSourceField)
        orderSql = "" if orderby is None else f"order by {orderby}"
        sql = f"select distinct {field._get_sql_field_name()} from {field.source_name} {orderSql}"
        cp = Slicer(SqlInfo(sql))
        cp.title = field.name
        cp.add_updateInfo(field.source_name, field._get_sql_field_name())

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        self.__resourceManager.register_element_cps()
        return cp

    def add_table(
        self,
        dataSourceTable: DataSourceTable,
        *,
        host: Optional[ContainerComponent] = None,
    ):
        sql = ""
        if dataSourceTable._user_specified_field:
            sql = dataSourceTable._to_sql()
        else:
            sql = f"select * from {dataSourceTable.source_name}"
        cp = Table(SqlInfo(sql))

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        self.__resourceManager.register_element_cps()
        return cp

    def add_echart(
        self,
        options: Union[Dict, BaseChart, ChartCollector],
        *,
        host: Optional[ContainerComponent] = None,
    ):
        cp = None

        if not isinstance(options, (BaseChart, ChartCollector, dict)):
            from pyecharts.charts.base import Base

            if isinstance(options, Base):
                options = options.get_options()
                assert isinstance(options, dict)
                pyecharts_utils.replace_jscode(options)
                #
                self.__json_utils.mark_pyecharts()

        if isinstance(options, (BaseChart, ChartCollector)):
            opts = options

            if isinstance(opts, BaseChart):
                opts = ChartCollector().append(opts)

            cp = EChart(option_type="dict")

            if (
                len(opts._collector) > 1
                and easy_echarts._settings.drill_down_default_set_click_filter
            ):
                for chart in opts._collector:
                    chart._create_default_click_filter()

            for chart in opts._collector:
                opts, updateInfos, mapIds = chart.get_options_infos()

                ds_infos = []
                jscodes = []

                OptionsExtractor.extract_and_remove_from_dict(opts, ds_infos, jscodes)

                ds_infos = [
                    EChartDatasetInfo({}, path, sql._sql_info) for path, sql in ds_infos
                ]

                jscodes = [EChartJscode(path, jscode) for path, jscode in jscodes]

                for map_name in mapIds:
                    self.__webResourceManager.mark_echarts_map(map_name)

                info = EChartInfo(opts, ds_infos, updateInfos, jscodes, mapIds)
                cp._add_chart_info(info)

        elif isinstance(options, dict):
            cp = EChart(option_type="dict")

            ds_infos = []
            jscodes = []

            OptionsExtractor.extract_and_remove_from_dict(options, ds_infos, jscodes)

            ds_infos = [
                EChartDatasetInfo({}, path, sql._sql_info) for path, sql in ds_infos
            ]

            jscodes = [EChartJscode(path, jscode) for path, jscode in jscodes]

            info = EChartInfo(options, ds_infos, [], jscodes)
            cp._add_chart_info(info)

        host = host or self._get_temp_host() or self

        assert cp is not None
        host._add_children(cp)

        self.__resourceManager.register_echarts()
        return cp

    def flowBox(
        self,
        align: Optional[str] = None,
        *,
        host: Optional[ContainerComponent] = None,
    ):
        """行容器，里面的组件会横向放置，在超过屏幕宽度时，自动换行

        Args:
            align (Optional[str], optional): 容器中的元素组件靠向哪个位置。'left' | 'center' | 'right' | 'between' | 'around' |'evenly' |'stretch'. Defaults to None('left').

                - left : 从行首起始位置开始排列
                - center:居中排列
                - right:从行尾位置开始排列
                - between: 均匀排列每个元素,首个元素放置于起点，末尾元素放置于终点
                - around:  均匀排列每个元素,每个元素周围分配相同的空间
                - evenly:  均匀排列每个元素, 每个元素之间的间隔相等

        ## 示例

        >>> # 2个文本放置在同一行，并且靠左放置
        >>> with pbi.flowBox():
                pbi.add_text('text A')
                pbi.add_text('text B')

        >>> # 2个文本放置在同一行，并且居中放置
        >>> with pbi.flowBox('center'):
                pbi.add_text('text A')
                pbi.add_text('text B')
        """
        cp = FlowBoxComponent(align=align, appHost=self)

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp

    def gridBox(
        self,
        areas: Union[List[List[str]], str],
        align: Optional[str] = None,
        vertical_align: Optional[str] = None,
        *,
        host: Optional[ContainerComponent] = None,
    ):
        """网格布局

        Args:
            areas (Union[List[List[str]], str]): 网格布局文本
        ---
        ### 示例
        ```python
        grid='''
            a b c
            a e .
        '''

        with pbi.gridBox(grid):
            pbi.add_slicer(...).set_gridArea('a')
            pbi.add_slicer(...).set_gridArea('b')
        ```


        当 `areas` 设置为 '' 时,默认为动态布局，每行放置3格
        ```python
        with pbi.gridBox(''):
            # 3个切片器在第1行
            pbi.add_slicer(...)
            pbi.add_slicer(...)
            pbi.add_slicer(...)

            # 2个切片器在第2行
            pbi.add_slicer(...)
            pbi.add_slicer(...)
        ```

        可以通过 `auto_fill_fixed_num` 修改动态配置

        >>> with pbi.gridBox('').auto_fill_fixed_num(2):
                ...

        """
        if isinstance(areas, str):
            areas = GridBoxComponent.areas_str2array(areas)

        cp = GridBoxComponent(areas, align, vertical_align, self)

        if len(areas) == 0:
            cp.auto_fill_by_num()

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp

    def colBox(
        self,
        spec: List[int] | None = None,
        *,
        host: Optional[ContainerComponent] = None,
    ):
        spec = spec or [1, 1]
        grid_text = " ".join(f"x{n}" for n in range(len(spec)))
        cp = self.gridBox(grid_text).set_columns_sizes(spec)

        return cp

    def box(
        self,
        align: Optional[str] = None,
        vertical_align: Optional[str] = None,
        *,
        host: Optional[ContainerComponent] = None,
    ):
        """盒容器，里面的组件竖向放置

        Args:
            align (Optional[str], optional): 容器中的元素组件横向靠哪个位置。
                'left' | 'center' | 'right' | 'between' | 'around' |'evenly'|'stretch'. Defaults to None('left').

                - left : 从行首起始位置开始排列
                - center:居中排列
                - right:从行尾位置开始排列

            vertical_align (Optional[str], optional): 容器中的元素垂直方向靠哪个位置。
                'top' | 'center' | 'bottom' | 'between' | 'around' . Defaults to None('top').

                - top : 靠顶部
                - center:居中
                - bottom:靠底部
                - between:项目在行与行之间留有间隔
                - around:项目在行之前、行之间和行之后留有空间

        ## 示例

        """
        cp = BoxComponent(align=align, vertical_align=vertical_align, appHost=self)
        host = host or self._get_temp_host() or self
        host._add_children(cp)
        return cp

    def add_tabs(
        self,
        names: List[str],
        *,
        host: Optional[ContainerComponent] = None,
    ):
        """ """
        cp = TabsComponent(names, appHost=self)
        host = host or self._get_temp_host() or self
        host._add_children(cp)

        self.__resourceManager.register_element_cps()
        return cp

    def save_zip_db(self, path: str):
        with open(path, mode="w", encoding="utf8") as f:
            f.write(ds2sqlite_file_base64(self.dataSources))

    def save_db(self, path: str):
        if Path(path).exists():
            os.remove(path)
        ds2sqlite(path, self.dataSources)

    def _to_json_dict(self):
        self.webResources.append(
            {
                "id": "DbFile",
                "type": "DbFile",
                "input": ds2sqlite_file_base64(
                    self.dataSources, clear_data=self._clear_data
                ),
                "actionPipe": [],
            }
        )

        wrs = self.__webResourceManager.create_webResources()
        self.webResources.extend(wrs)

        data = super()._to_json_dict()

        if self._doc_title:
            data["docTitle"] = self._doc_title

        if self._style_tags:
            data["styleTags"] = self._style_tags

        data["version"] = pbi.__version__
        return data

    def __reset_data(self):
        """support for run on ipython env"""
        self.children = []
        self.dataSources = []
        self.dataViews = []

    def to_json(self, *args, **kws):
        return self.__json_utils.dumps(self, *args, **kws)
        # return json_dumps_fn(self, indent=2, ensure_ascii=False)

    def to_raw_html(self):
        try:
            symbol = '"__{{__config_data__}}___"'

            config = self.__json_utils.dumps(self)

            with open(
                get_project_root() / "template/index.html", mode="r", encoding="utf8"
            ) as html:
                res = html.read().replace(symbol, config)
                return res
        except Exception as e:
            raise e
        else:
            self.__reset_data()

    def to_html(self, file, display_output_path=False):
        try:
            # file = Path(file)
            # raw = self.to_raw_html()

            raw = self.__resourceManager.build_html(
                self.to_json(), svg_infos=get_iconManager().get_infos()
            )
            Path(file).write_text(raw, "utf8")

            if display_output_path:
                print(f"to html:{file.absolute()}")
        except Exception as e:
            raise e
        else:
            self.__reset_data()

    def add_markdown(
        self,
        md: str,
        *,
        host: Optional[ContainerComponent] = None,
    ):
        # if isinstance(md, SqlWrapper):
        #     md = str(md)
        md = SqlInfo.around_backticks(md)

        html = markdown(
            md,
            extras=[
                "fenced-code-blocks",
                "target-blank-links",
                "task_list",
                "code-color",
                "tag-friendly",
            ],
        )

        contents = list(SqlInfo.extract_sql_from_text(html))

        cp = Markdown(contents)

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp

    def add_icon(
        self,
        icon_svg: str,
        size="2em",
        color="currentColor",
        *,
        host: Optional[ContainerComponent] = None,
    ):
        icon_id = get_iconManager().make_icon(icon_svg)
        cp = IconComponent(icon_id, size, color)

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp

    def add_svg_icon(
        self,
        svg: str,
        size="2em",
        color="currentColor",
        *,
        host: Optional[ContainerComponent] = None,
    ):
        cp = SvgIconComponent(svg, size, color)

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp

    def affix(
        self,
        *,
        host: Optional[ContainerComponent] = None,
    ):
        cp = AffixComponent(self)
        host = host or self._get_temp_host() or self
        host._add_children(cp)

        self.__resourceManager.register_element_cps()
        return cp

    def add_mermaid(
        self,
        graph: str,
        *,
        host: Optional[ContainerComponent] = None,
    ):
        """
        [mermaid文档](https://github.com/mermaid-js/mermaid/blob/develop/README.zh-CN.md)
        ---
        >>> graph = '''
        flowchart LR
        A[Hard] -->|Text| B(Round)
        B --> C{Decision}
        C -->|One| D[Result 1]
        C -->|Two| E[Result 2]
        '''
        >>> add_mermaid(graph)
        """
        cp = Mermaid(graph)
        host = host or self._get_temp_host() or self
        host._add_children(cp)

        self.__resourceManager.register_mermaid_cps()
        return cp

    def add_input(
        self,
        field: Union[DataSourceField, DataSourceTable],
        where_expr="like '%${}%'",
        *,
        host: Optional[ContainerComponent] = None,
    ):
        """ """
        if isinstance(field, DataSourceTable):
            field = field[field.columns[0]]

        assert isinstance(field, DataSourceField)

        cp = Input(where_expr).add_updateInfo(field.source_name, field.name)
        host = host or self._get_temp_host() or self
        host._add_children(cp)

        self.__resourceManager.register_element_cps()
        return cp

    def add_numberSlider(
        self,
        data: DataSourceTable,
        field: str,
        where_expr=" between ${0} and ${1}",
        range=True,
        *,
        host: Optional[ContainerComponent] = None,
    ):
        """ """
        cp = (
            NumberSlider(where_expr)
            .add_updateInfo(data.source_name, field)
            .set_props({"range": range})
        )
        host = host or self._get_temp_host() or self
        host._add_children(cp)

        self.__resourceManager.register_element_cps()
        return cp

    def space(
        self,
        space: Optional[str] = None,
        *,
        auto_fill_row=False,
        host: Optional[ContainerComponent] = None,
    ):
        """增加组件之间的空间

        Args:
            space (Optional[str], optional): 间距值(默认为'0.2em'). Defaults to None.
            auto_fill_row (bool, optional): 放在 flowBox 中,会迫使后续的组件往右靠. Defaults to False.
        ---
        >>> pbi.space()
        >>> pbi.space('100px')
        >>> pbi.space('0.5em')
        """
        cp = SpaceComponent(space, auto_fill_row)
        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp



from __future__ import annotations


from .base import BaseChart
import pybi as pbi


from typing import TYPE_CHECKING, Optional
from . import utils as easyUtils

if TYPE_CHECKING:
    from pybi.core.dataSource import DataSourceTable


class CandleStickChart(BaseChart):
    def __init__(
        self,
        data: DataSourceTable,
        date="date",
        open="open",
        close="close",
        lowest="lowest",
        highest="highest",
    ):
        super().__init__()
        self.data = data
        self.date = date
        self.open = open
        self.close = close
        self.lowest = lowest
        self.open = open
        self.highest = highest
        self._series_configs = {}

    def _create_default_click_filter(self):
        pass

    def _create_options_ex(self):
        opts = super().get_options()

        base_opt = {
            "xAxis": [{}],
            "yAxis": [{}],
            "series": [],
            "dataZoom": [
                {
                    "textStyle": {"color": "#8392A5"},
                    "handleIcon": "path://M10.7,11.9v-1.3H9.3v1.3c-4.9,0.3-8.8,4.4-8.8,9.4c0,5,3.9,9.1,8.8,9.4v1.3h1.3v-1.3c4.9-0.3,8.8-4.4,8.8-9.4C19.5,16.3,15.6,12.2,10.7,11.9z M13.3,24.4H6.7V23h6.6V24.4z M13.3,19.6H6.7v-1.4h6.6V19.6z",
                    "dataBackground": {
                        "areaStyle": {"color": "#8392A5"},
                        "lineStyle": {"opacity": 0.8, "color": "#8392A5"},
                    },
                    "brushSelect": True,
                },
                {"type": "inside"},
            ],
            "tooltip": {
                "trigger": "axis",
                "axisPointer": {"type": "cross"},
                "borderWidth": 1,
                "borderColor": "#ccc",
                "padding": 10,
                "textStyle": {"color": "#000"},
            },
        }

        # base_opt = easyUtils.merge_user_options(base_opt, opts)
        base_opt.update(opts)

        base_opt["series"].append(
            {
                "type": "candlestick",
                "data": pbi.sql(
                    f"select {self.open},{self.close},{self.lowest},{self.highest}  from {self.data} order by {self.date}"
                ).toflatlist(),
                "universalTransition": {"enabled": True, "divideShape": "clone"},
            }
        )

        base_opt["xAxis"][0]["data"] = pbi.sql(
            f"select distinct {self.date} from {self.data} order by {self.date}"
        ).toflatlist()

        return base_opt, self._updateInfos, []



from __future__ import annotations
from .line import LineChart
from .bar import BarChart
from .pie import PieChart
from .scatter import ScatterChart
from .map import MapChart
from .candlestick import CandleStickChart
from .radar import RadarChart
from typing import TYPE_CHECKING, Optional
import pybi as pbi

if TYPE_CHECKING:
    from pybi.core.dataSource import DataSourceTable


__all__ = ["easy_echarts"]


class EasyEChartsSettings:
    def __init__(self) -> None:
        self.drill_down_default_set_click_filter = True


class EasyEChartsMeta:
    def __init__(self) -> None:
        self._settings = EasyEChartsSettings()

    def off_drill_down_default_set_click_filter(self):
        """
        默认情况下,多个组合图表配置会自动为其每个图表添加 `click_filter` 联动。
        此函数调用后,会关闭其功能
        """
        self._settings.drill_down_default_set_click_filter = False

    def make_line(
        self,
        data: DataSourceTable,
        x: str,
        y: str,
        color: Optional[str] = None,
        agg="round(avg(${}),2)",
        order: Optional[str] = None,
    ):
        order = order or x
        return LineChart(data, x, y, order, color, agg)

    def make_bar(
        self,
        data: DataSourceTable,
        *,
        x: str,
        y: str,
        color: Optional[str] = None,
        agg="round(avg(${}),2)",
        order: Optional[str] = None,
    ):
        order = order or x
        return BarChart(data, x, y, order, color, agg)

    def make_pie(
        self,
        data: DataSourceTable,
        *,
        name: str,
        value: str,
        agg="round(avg(${}),2)",
    ):
        return PieChart(data, name, value, agg)

    def make_scatter(
        self,
        data: DataSourceTable,
        *,
        x: str,
        y: str,
        color: Optional[str] = None,
        agg="round(avg(${}),2)",
    ):
        return ScatterChart(data, x, y, color, agg)

    def make_map(self, level="province"):
        return MapChart(level)

    def make_candleStick(
        self,
        data: DataSourceTable,
        *,
        date="date",
        open="open",
        close="close",
        lowest="lowest",
        highest="highest",
    ):
        """
        candleStick 即我们常说的 K线图
        [date,open, close, lowest, highest] (即：[日期, 开盘值, 收盘值, 最低值, 最高值])
        """
        return CandleStickChart(data, date, open, close, lowest, highest)

    def make_radar(
        self,
        data: DataSourceTable,
        *,
        indicator="indicator",
        name="name",
        value="value",
        agg="round(avg(${}),2)",
    ):
        """
        雷达图
        ---
        indicator:雷达图的指示器，用来指定雷达图中的多个变量（维度）
        name:数据项名称
        value:具体的数据，每个值跟 indicator 一一对应。
        ---
        示例：
        ```python
        data = pd.DataFrame(
            {
                "指标": ["A", "B", "C", "D", "E", "F", "A", "B", "C", "D", "F"],
                "值": [8150, 1035, 3393, 7919, 4244, 6237, 3537, 5676, 8306, 6388, 3889],
                "类别": ["X", "X", "X", "X", "X", "X", "Y", "Y", "Y", "Y", "Y"],
            }
        )

        data = pbi.set_source(data)
        pbi.add_table(data)
        opts = charts.make_radar(data, indicator="指标", name="类别", value="值")
        pbi.add_echart(opts).set_height("30em")
        ```
        """
        return RadarChart(data, indicator, name, value, agg)


easy_echarts = EasyEChartsMeta()



from .base import *
from .echarts import *
from .slicer import *
from .table import *
from .textValue import *
from .markdown import Markdown
from .input import Input
from .numberSlider import NumberSlider



import base64
import datetime
from enum import Enum
from pathlib import Path
import simplejson as json
from functools import partial
from typing import Callable, Sequence, List, Dict
import numpy as np
import pandas as pd
import inspect

from abc import abstractmethod
import pybi.utils.echarts_opts_utils as echarts_opts_utils
import pybi.utils.pyecharts_utils as pybi_pyecharts_utils

_global_id = 0


class Jsonable:
    @abstractmethod
    def _to_json_dict(self):
        data = {k: v for k, v in self.__dict__.items() if k[:1] != "_"}

        return data


def fn2str(fn: Callable):
    return inspect.getsource(fn)


def random_ds_name():
    return f"ds_{get_global_id()}"


def random_dv_name():
    return f"dv_{get_global_id()}"


def get_global_id():
    global _global_id
    _global_id += 1
    return str(_global_id)


_m_project_root = Path(__file__).absolute().parent.parent


def get_project_root():
    return _m_project_root


def pybi_json_default(obj):
    if isinstance(obj, float) and pd.isna(obj):
        return None
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    if isinstance(obj, (datetime.date, datetime.datetime)):
        return obj.isoformat()
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, Jsonable):
        return obj._to_json_dict()


class JsonUtils:
    def __init__(self) -> None:
        self.json_defalut_fns: List[Callable] = [pybi_json_default]

    def mark_pyecharts(self):
        if len(self.json_defalut_fns) >= 2:
            return

        from pyecharts.options.series_options import BasicOpts
        from pyecharts.commons import utils as pyecharts_utils

        self.json_defalut_fns.append(
            partial(
                pybi_pyecharts_utils.pyecharts_json_default,
                utils=pyecharts_utils,
                basicOptsType=BasicOpts,
            )
        )

    def dumps(self, obj, *args, **kws):
        def json_default(obj):
            for fn in self.json_defalut_fns:
                res = fn(obj)
                if res is not None:
                    return res

        return json.dumps(
            obj, *args, **kws, ignore_nan=True, default=json_default, ensure_ascii=False
        )


# json_dumps_fn = partial(json.dumps, ignore_nan=True, default=json_default)


def data2html_img_src(data: bytes):
    b64 = base64.b64encode(data).decode("utf8")
    value = f"data:image/png;base64,{b64}"
    return value


def file2html_base64_src(file: str, format: str):

    with open(file, mode="rb") as f:
        b64 = base64.b64encode(f.read())
        b64 = str(b64, "utf8")
        value = f"data:{format};base64,{b64}"
        return value


class StrEnum(str, Enum):
    def __new__(cls, value, *args, **kwargs):
        return super().__new__(cls, value, *args, **kwargs)

    def __str__(self):
        return str(self.value)

    def _generate_next_value_(name, *_):
        return name



from __future__ import annotations
from typing import TYPE_CHECKING, List, Union


from pybi.core.components import ComponentTag
from .base import SingleReactiveComponent


if TYPE_CHECKING:
    from pybi.core.sql import SqlInfo


class Table(SingleReactiveComponent):
    def __init__(
        self,
        sql: SqlInfo,
    ) -> None:
        super().__init__(ComponentTag.Table, sql)
        self.pageSize = 10
        self.tableHeight = "initial"
        self.tableWidth = "initial"

    def set_page_size(self, size: int):
        """
        设置表格每页行数
        size: >=5 ,默认10
        """
        self.pageSize = max(size, 5)
        return self

    def set_table_height(self, height="initial"):
        """
        表格高度
        height: 'initial'(默认值),'30em','30%','30vh'
        如果设置为initial,则表格会以展示一页所有数据的高度作为固定高度
        """
        self.tableHeight = height
        return self

    def set_table_width(self, width="initial"):
        """
        表格高度
        width: 'initial'(默认值),'30em','30%','30vh'
        """
        self.tableWidth = width
        return self



from __future__ import annotations
from typing import TYPE_CHECKING, Dict, Union, List, Optional

from .componentTag import ComponentTag
from .component import Component
from pybi.icons.iconManager import get_singleton as get_iconManager

if TYPE_CHECKING:
    from pybi.app import App


class ContainerComponent(Component):
    def __init__(
        self,
        tag: ComponentTag,
        appHost: Optional[App] = None,
        childredHook: Optional[ContainerComponent] = None,
    ) -> None:
        super().__init__(tag)

        # TODO: maybe use weak ref?
        self._appHost = appHost

        self._childredHook = childredHook
        self.children: List[Component] = []

    def _add_children(self, stat: Component):
        if self._childredHook:
            self._childredHook._add_children(stat)
        else:
            self.children.append(stat)
        return self

    def __enter__(self):
        if self._appHost:
            self._appHost._with_temp_host_stack.append(self)
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        if self._appHost:
            self._appHost._with_temp_host_stack.pop()


class BoxComponent(ContainerComponent):
    m_align_mapping = {
        "left": "flex-start",
        "center": "center",
        "right": "flex-end",
        "stretch": "stretch",
    }

    m_vertical_align_mapping = {
        "top": "flex-start",
        "center": "center",
        "bottom": "flex-end",
        "between": "space-between",
        "around": "space-around",
    }

    def __init__(
        self,
        align: Optional[str] = None,
        vertical_align: Optional[str] = None,
        appHost: Optional[App] = None,
    ) -> None:
        """
        Args:
            align (Optional[str], optional): 'left' | 'center' | 'right'. Defaults to None.
            vertical_align (Optional[str], optional): 'top' | 'center' | 'bottom'. Defaults to None.
        """
        super().__init__(ComponentTag.Box, appHost)
        self._align = align
        self._vertical_align = vertical_align

    def _to_json_dict(self):
        data = super()._to_json_dict()
        if self._align:
            data["align"] = self.m_align_mapping[self._align]

        if self._vertical_align:
            data["verticalAlign"] = self.m_vertical_align_mapping[self._vertical_align]

        return data


class FlowBoxComponent(ContainerComponent):
    m_align_mapping = {
        "left": "flex-start",
        "center": "center",
        "right": "flex-end",
        "between": "space-between",
        "around": "space-around",
        "evenly": "space-evenly",
        "stretch": "stretch",
    }

    def __init__(
        self, align: Optional[str] = None, appHost: Optional[App] = None
    ) -> None:
        """
        Args:
            align (Optional[str], optional): 'left' | 'center' | 'right'. Defaults to None.
        """
        super().__init__(ComponentTag.FlowBox, appHost)
        self._align = align

    def __get_item__(self, idx: int):
        return self.children[idx]

    def _to_json_dict(self):
        data = super()._to_json_dict()
        if self._align:
            data["align"] = self.m_align_mapping[self._align]

        return data


class GridBoxComponent(ContainerComponent):
    m_align_mapping = {
        "left": "flex-start",
        "center": "center",
        "right": "flex-end",
        "stretch": "stretch",
    }

    m_vertical_align_mapping = {
        "top": "flex-start",
        "center": "center",
        "bottom": "flex-end",
        "between": "space-between",
        "around": "space-around",
    }

    def __init__(
        self,
        areas: Optional[List[List[str]]] = None,
        align: Optional[str] = None,
        vertical_align: Optional[str] = None,
        appHost: Optional[App] = None,
    ) -> None:
        super().__init__(ComponentTag.GridBox, appHost)
        self.__areas = areas
        self.__columns_sizes: List[str] = []
        self.__rows_sizes: List[str] = []
        self.__gridTemplateColumns = None

        self._align = align
        self._vertical_align = vertical_align

    def set_columns_sizes(self, sizes: List[Union[str, int]]):
        """
        >>> set_columns_sizes([1,2,2])
        >>> set_columns_sizes(['150px','1fr','2fr'])
        >>> set_columns_sizes(['150px','1fr','minmax(100px, 1fr)'])
        """
        self.__columns_sizes = [f"{s}fr" if isinstance(s, int) else s for s in sizes]
        return self

    def set_rows_sizes(self, sizes: List[Union[str, int]]):
        """
        >>> set_rows_sizes([1,2,2])
        >>> set_rows_sizes(['150px','1fr','2fr'])
        >>> set_rows_sizes(['150px','1fr','minmax(100px, 1fr)'])
        """
        self.__rows_sizes = [f"{s}fr" if isinstance(s, int) else s for s in sizes]
        return self

    def __get_item__(self, idx: int):
        return self.children[idx]

    def auto_fill_by_width(self, width="15em"):
        """动态调整每行组件数量，每个组件固定宽度为指定 `width`。
        当设置有效 grid areas 时，此设置无效

        Args:
            width (str, optional): 每个组件固定宽度. Defaults to "15em".

        """

        width_value = f"minmax(0, {width})"
        self.__gridTemplateColumns = f"repeat(auto-fill, {width_value})"
        self.set_style("justify-content: space-around")
        return self

    def auto_fill_by_num(self, num=3):
        """自动调整每行组件宽度，每行组件数量为 `num`。
        当设置有效 grid areas 时，此设置无效

        Args:
            num (int, optional): 每行组件数量. Defaults to 3.

        """
        self.__gridTemplateColumns = f"repeat({num}, 1fr)"
        return self

    def _to_json_dict(self):
        data = super()._to_json_dict()

        if self.__areas:
            data["areas"] = GridBoxComponent.areas_array2str(self.__areas)

            areas_cols_len = max(map(len, self.__areas))
            cols_size = GridBoxComponent.padded_grid_template(
                self.__columns_sizes, areas_cols_len, "1fr"
            )

            areas_rows_len = len(self.__areas)
            rows_size = GridBoxComponent.padded_grid_template(
                self.__rows_sizes, areas_rows_len
            )

            data["gridTemplateColumns"] = " ".join(cols_size)
            data["gridTemplateRows"] = " ".join(rows_size)
        else:
            if self.__gridTemplateColumns:
                data["gridTemplateColumns"] = self.__gridTemplateColumns

        if self._align:
            data["align"] = self.m_align_mapping[self._align]

        if self._vertical_align:
            data["verticalAlign"] = self.m_vertical_align_mapping[self._vertical_align]

        return data

    @staticmethod
    def padded_grid_template(sizes: List[str], real_size: int, pad="auto"):
        """
        >>> sizes = ['1fr']
        >>> real_size = 3
        >>> padded_grid_template(sizes,real_size)
        >>> ['1fr','auto','auto']
        """
        diff_len = real_size - len(sizes)
        sizes = sizes.copy()

        if diff_len > 0:
            sizes.extend([pad] * diff_len)
        return sizes

    @staticmethod
    def areas_array2str(areas_array: List[List[str]]):
        """
        >>> input = [
            ["sc1", "sc2"],
            ["sc3"],
            ["table"] * 4
        ]
        >>> areas_array2str(input)
        >>> '"sc1 sc2 . ." "sc3 . . ." "table table table table"'
        """
        max_len = max(map(len, areas_array))

        fix_empty = (
            [*line, *(["."] * (max_len - len(line)))] if len(line) < max_len else line
            for line in areas_array
        )

        line2str = (f'"{" ".join(line)}"' for line in fix_empty)
        return " ".join(line2str)

    @staticmethod
    def areas_str2array(areas: str) -> List[List[str]]:
        """
        >>> input='''
            sc1 sc2
            sc3
            table table table table
        '''
        >>> areas_str2array(input)
        >>> [
            ["sc1", "sc2"],
            ["sc3"],
            ["table", "table", "table", "table"]
        ]
        """
        pass

        lines = (line.strip() for line in areas.splitlines())
        remove_empty_rows = (line for line in lines if len(line) > 0)
        splie_space = (line.split() for line in remove_empty_rows)
        return list(splie_space)


class TabsComponent(ContainerComponent):
    def __init__(self, names: List[str], appHost: Optional[App] = None) -> None:
        """
        mode: 'fullWidth' | 'narrowing'
        """
        super().__init__(ComponentTag.Tabs, appHost)
        if len(names) > len(set(names)):
            raise Exception("names cannot be duplicated")
        self.names = names
        self.__icons = []
        self.tabsProps = {}
        self.panelsProps = {}
        self.tabsClasses: List[str] = []
        self.panelsClasses: List[str] = []

        self.__name2idx = {name: idx for idx, name in enumerate(self.names)}

        for _ in range(len(self.names)):
            self._add_children(BoxComponent(appHost=appHost))

    def __getitem__(self, idx: Union[int, str]):
        if isinstance(idx, str):
            if idx not in self.__name2idx:
                raise Exception(f"tab name[{idx}] not found")
            idx = self.__name2idx[idx]

        res = self.children[idx]
        assert isinstance(res, ContainerComponent)
        return res

    def set_icons(self, icons: List[str]):
        self.__icons = icons

        self.set_tabsProps({"inline-label": len(icons) > 0})
        return self

    def set_tabsProps(self, props: Dict):
        self.tabsProps.update(props)
        return self

    def set_panelsProps(self, props: Dict):
        self.panelsProps.update(props)
        return self

    def set_tabsClasses(self, value: str):
        """
        set_tabsClasses('text-primary bg-positive')
        """
        values = (v for v in value.split(" ") if v)
        self.tabsClasses.extend(values)
        return self

    def set_panelsClasses(self, value: str):
        """
        set_panelsClasses('text-primary bg-positive')
        """
        values = (v for v in value.split(" ") if v)
        self.panelsClasses.extend(values)
        return self

    def set_props(self, props: Dict):
        return self.set_tabsProps(props)

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["tabsClasses"] = list(dict.fromkeys(data["tabsClasses"]).keys())
        data["panelsClasses"] = list(dict.fromkeys(data["panelsClasses"]).keys())

        icon_ids = []
        for icon in self.__icons:
            icon_id = get_iconManager().make_icon(icon)
            icon_ids.append(icon_id)

        data["iconIds"] = icon_ids

        return data


class AffixComponent(ContainerComponent):
    def __init__(self, appHost: Optional[App] = None) -> None:
        super().__init__(ComponentTag.Affix, appHost)
        box = BoxComponent(appHost=appHost)
        self._add_children(box)
        self._childredHook = box


class DynamicBoxComponent:
    def __init__(self, grid_box: GridBoxComponent) -> None:
        self.__grid_box = grid_box

    def set_each_row_num(self, num: int):
        pass

    def set_each_width(self, width: str):
        pass



from typing import Any, Dict


def set_prop_by_path(options: Dict, path: str, value: Any):
    """
    path: 'sreies[0].data[0].item.value'
    """
    paths = [p for p in path.split(".") if len(p) > 0]
    target = options
    for path in paths[:-1]:
        if path[-1] == "]":
            # e.g sreies[0]
            start_idx = path.index("[")
            key = path[:start_idx]
            path_idx = int(path[start_idx + 1 : -1])
            target = target[key][path_idx]
        else:
            # e.g .item
            if path not in target:
                target[path] = {}
            target = target[path]

    path = paths[-1]
    if path[-1] == "]":
        # e.g sreies[0]
        start_idx = path.index("[")
        key = path[:start_idx]
        path_idx = int(path[start_idx + 1 : -1])
        target[key][path_idx] = value
    else:
        # e.g .item
        target[path] = value


def iter_each_items(options: Dict):
    """
    >>> opts = {'series':[{'type':'bar',data:[1,2]}]}
    >>> target,path = next(iter_each_items(opts))
    >>> target,path
    >>> [{'type':'bar',data:[1,2]}] , '.series'
    >>> {'type':'bar',data:[1,2]} , '.series[0]'
    """
    stack = [(options, "")]

    def interception_type(target, path):
        if isinstance(target, dict):
            inputs = ((value, f"{path}.{key}") for key, value in target.items())
            stack.extend(inputs)
            return True

        if isinstance(target, list):
            inputs = ((value, f"{path}[{idx}]") for idx, value in enumerate(target))
            stack.extend(inputs)
            return True

        return False

    while len(stack) > 0:
        target, path = stack.pop()

        if interception_type(target, path):
            continue

        resend = yield target, path
        if resend is not None:
            interception_type(resend, path)



from __future__ import annotations
from .base import BaseChart
import pybi as pbi
import pybi.utils.sql as sqlUtils

from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from pybi.core.dataSource import DataSourceTable


class PieChart(BaseChart):
    def __init__(
        self,
        data: DataSourceTable,
        name: str,
        value: str,
        agg="round(avg(${}),2)",
    ):
        super().__init__()
        self.data = data
        self.name = name
        self.value = value
        self.agg = agg
        self._series_configs = {}

    def radius(self, value=["40%", "70%"]):
        """
        设置圆环图颜色。
        参考资料:https://echarts.apache.org/zh/option.html#series-pie.radius
        """
        self._series_configs["radius"] = value
        self._series_configs["avoidLabelOverlap"] = False
        self._series_configs["label"] = {"show": False, "position": "center"}
        self._series_configs["labelLine"] = {"show": False}
        self._series_configs["emphasis"] = {
            "label": {"show": True, "fontSize": "2em", "fontWeight": "bold"}
        }

        return self

    def _create_default_click_filter(self):
        self.click_filter("name", self.data, self.name)

    def _create_options_ex(self):
        opts = super().get_options()

        base_opt = {
            "xAxis": None,
            "yAxis": None,
            "series": [],
        }
        base_opt.update(opts)

        series_config = {
            "id": self.name,
            "universalTransition": {"enabled": True, "divideShape": "clone"},
            "type": "pie",
            "radius": "50%",
            "emphasis": {
                "itemStyle": {
                    "shadowBlur": 10,
                    "shadowOffsetX": 0,
                    "shadowColor": "rgba(0, 0, 0, 0.5)",
                }
            },
        }
        series_config.update(self._series_configs)

        agg_field = sqlUtils.apply_agg(self.agg, self.value)
        sql = pbi.sql(
            f"select `{self.name}` as name ,{agg_field} as value,`{self.name}` as groupId from {self.data} group by `{self.name}`"
        )

        series_config["data"] = sql

        base_opt["series"].append(series_config)

        opts["tooltip"] = {"trigger": "item"}

        return base_opt, self._updateInfos, []



class IconManager:
    def __init__(self) -> None:
        self.__id_number = 0
        self.__mapping_svg_id = {}

    def _create_id(self):
        id = f"icon_{self.__id_number}"
        self.__id_number = self.__id_number + 1
        return id

    def make_icon(self, svg_content: str):
        if not svg_content in self.__mapping_svg_id:
            self.__mapping_svg_id[svg_content] = self._create_id()

        return self.__mapping_svg_id[svg_content]

    def get_infos(self):
        infos = [{"id": id, "svg": svg} for svg, id in self.__mapping_svg_id.items()]

        return infos


__singleton = IconManager()


def get_singleton():
    return __singleton



from pybi.core.components.reactiveComponent.echarts import EChartJsCode
from .app import App

# import pybi.core.styles as styles
from pybi.core.styles import *
from pybi.easyEcharts import *
import pybi.icons.material_icons as md_icons

md_icons = md_icons
"""material icons小图标
    >>> pbi.add_icon(pbi.md_icons.md_add_chart)
"""

app = App()

__all__ = [
    "set_source",
    "add_slicer",
    "add_table",
    "add_echart",
    "add_text",
    "colBox",
    "flowBox",
    "gridBox",
    "box",
    "app",
    "styles",
    "easy_echarts",
    "echartJsCode",
    "to_json",
    "sql",
    "set_dataView",
    "to_html",
    "clear_all_data",
    "add_upload",
    "_save_db",
    "meta",
    "save_zip_db",
    "add_tabs",
    "add_markdown",
    "add_icon",
    "affix",
    "add_mermaid",
    "add_input",
    "add_numberSlider",
    # "add_svg_icon",
    "md_icons",
    "space",
]

add_tabs = app.add_tabs
meta = app.meta
gridBox = app.gridBox
set_source = app.set_source
add_upload = app.add_upload
add_text = app.add_text
add_slicer = app.add_slicer
add_table = app.add_table
add_echart = app.add_echart
colBox = app.colBox
flowBox = app.flowBox
echartJsCode = EChartJsCode
box = app.box
to_json = app.to_json
set_dataView = app.set_dataView
sql = app.sql
clear_all_data = app.clear_all_data
_save_db = app.save_db
save_zip_db = app.save_zip_db

to_html = app.to_html
add_markdown = app.add_markdown
add_icon = app.add_icon
# add_svg_icon = app.add_svg_icon
affix = app.affix
add_mermaid = app.add_mermaid
add_input = app.add_input
add_numberSlider = app.add_numberSlider
space = app.space



from __future__ import annotations
from typing import TYPE_CHECKING, List, Optional, Dict, Set, Any, Union
from pybi.core.sql import SqlWrapper

from pybi.utils.data_gen import Jsonable
from pybi.core.components import ComponentTag
from .base import ReactiveComponent
from pybi.core.dataSource import DataSourceTable
import pybi.utils.echarts_opts_utils as echarts_opts_utils

if TYPE_CHECKING:
    from pybi.core.dataSource import DataSourceTable
    from pybi.core.sql import SqlInfo


class EChartSqlInfo(Jsonable):
    def __init__(
        self, seriesConfig: Dict, path: str, sql: SqlInfo, chartType: str
    ) -> None:
        self.path = path
        self._sql = sql
        self.chartType = chartType
        self.seriesConfig = seriesConfig

    def _to_json_dict(self):
        data = super()._to_json_dict()
        sql_data = self._sql._to_json_dict()

        data = {**data, **sql_data}
        return data


class EChartDatasetInfo(Jsonable):
    def __init__(self, seriesConfig: Dict, path: str, sql_info: SqlInfo) -> None:
        self.path = path
        self.sqlInfo = sql_info
        self.seriesConfig = seriesConfig


class EChartJscode(Jsonable):
    def __init__(self, path: str, code: str) -> None:
        self.path = path
        self.code = code


class EChartUpdateInfo(Jsonable):
    def __init__(
        self,
        action_type: str,
        value_type: str,
        table: str,
        field: str,
    ) -> None:
        """
        action_type : Literal["hover", "click"]
        value_type: Literal["x", "y", "value","color","name"]
        """
        super().__init__()
        self.actionType = action_type
        self.valueType = value_type
        self.table = table
        self.field = field


class EChartJsCode:
    js_code_flag = "--x_x--0_0--"

    def __init__(self, code: str) -> None:
        self.code = code

    def __str__(self) -> str:
        return f"{EChartJsCode.js_code_flag}{self.code}{EChartJsCode.js_code_flag}"


class OptionsExtractor:
    @staticmethod
    def extract_and_remove_from_dict(
        data: Dict,
        out_datasets: List[tuple[str, SqlWrapper]],
        out_jscodes: List[tuple[str, str]],
    ):
        for target, path in echarts_opts_utils.iter_each_items(data):
            if isinstance(target, SqlWrapper):
                rpath = path[1:]
                out_datasets.append((rpath, target))
                OptionsExtractor.set_none_prop(data, rpath)

            if isinstance(target, EChartJsCode):
                rpath = path[1:]
                out_jscodes.append((rpath, target.code))
                OptionsExtractor.set_none_prop(data, rpath)

            if (
                isinstance(target, str)
                and target[: len(EChartJsCode.js_code_flag)]
                == EChartJsCode.js_code_flag
                and target[-len(EChartJsCode.js_code_flag) :]
                == EChartJsCode.js_code_flag
            ):
                target = target[
                    len(EChartJsCode.js_code_flag) : -len(EChartJsCode.js_code_flag)
                ]
                rpath = path[1:]
                out_jscodes.append((rpath, target))
                OptionsExtractor.set_none_prop(data, rpath)

    @staticmethod
    def set_none_prop(options: Dict, path: str):
        """
        path: 'sreies[0].data[0].item.value'
        """
        echarts_opts_utils.set_prop_by_path(options, path, None)


class EChartInfo(Jsonable):
    def __init__(
        self,
        options: Dict,
        datasetInfos: List[EChartDatasetInfo],
        updateInfos: List[EChartUpdateInfo],
        jscodes: List[EChartJscode],
        mapIds: Optional[List[str]] = None,
    ):
        self.options = options
        self.datasetInfos = datasetInfos
        self.updateInfos = updateInfos
        self.jsCodes = jscodes
        self.mapIds = mapIds or []


class EChart(ReactiveComponent):
    def __init__(self, *, option_type="dict") -> None:
        super().__init__(ComponentTag.EChart)
        self._chart_mappings = {}
        self._chartInfos: List[EChartInfo] = []
        self.optionType = option_type
        self.height = "100%"

    def set_height(self, value: str):
        """
        15em:15字体大小
        300px:300像素
        """
        self.height = value
        return self

    def _add_chart_info(self, info: EChartInfo):
        self._chartInfos.append(info)
        return self

    def hover_filter(
        self, value_type: str, table: Union[str, DataSourceTable], field: str
    ):
        """
        value_type: , Literal["x", "y", "value","color","name"]
        """
        if isinstance(table, DataSourceTable):
            table = table.source_name

        self._chartInfos[0].updateInfos.append(
            EChartUpdateInfo("hover", value_type, table, field)
        )

        return self

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["chartInfos"] = self._chartInfos
        return data



from enum import Enum


class ComponentTag(Enum):
    App = "App"
    DataSource = "DataSource"
    Slicer = "elSlicer"
    Input = "elInput"
    NumberSlider = "elNumberSlider"
    Table = "elTable"
    EChart = "EChart"
    Box = "Box"
    Text = "Text"
    Upload = "Upload"
    TextValue = "TextValue"
    Markdown = "Markdown"
    ColBox = "ColBox"
    FlowBox = "FlowBox"
    GridBox = "GridBox"
    Tabs = "elTabs"
    Icon = "Icon"
    SvgIcon = "SvgIcon"
    Affix = "Affix"
    Mermaid = "Mermaid"
    Space = "Space"



from __future__ import annotations
from typing import TYPE_CHECKING, Union, List, Callable, Dict

if TYPE_CHECKING:
    from pybi.core.components.component import Component


class Style:
    def __init__(self) -> None:
        self._dict_fn: List[Callable] = []

    def _copyFrom(self, other: Style):
        self._dict_fn.extend(other._dict_fn)
        return self

    def __add__(self, other: Style):
        new = Style()
        new._copyFrom(self)
        new._copyFrom(other)
        return new

    def _get_styles_dict(self):
        data = {}
        for fn in self._dict_fn:
            data.update(fn())

        return data


class StyleBuilder(Style):
    def __init__(self, data_dict: Dict) -> None:
        super().__init__()

        def fn():
            return data_dict

        self._dict_fn.append(fn)


class width(Style):
    def __init__(self, value: Union[str, int]) -> None:
        super().__init__()

        def fn():
            return {"width": value}

        self._dict_fn.append(fn)


class border(Style):
    """https://developer.mozilla.org/zh-CN/docs/Web/CSS/border"""

    def __init__(self, value="1px solid #000") -> None:
        super().__init__()

        def fn():
            return {"border": value}

        self._dict_fn.append(fn)


class borderRadius(Style):
    """https://developer.mozilla.org/zh-CN/docs/Web/CSS/border-radius"""

    def __init__(self, value="0.25rem") -> None:
        super().__init__()

        def fn():
            return {"border-radius": value}

        self._dict_fn.append(fn)


class padding(Style):
    """https://developer.mozilla.org/zh-CN/docs/Web/CSS/padding"""

    def __init__(self, value="0.8em") -> None:
        super().__init__()

        def fn():
            return {"padding": value}

        self._dict_fn.append(fn)


class textAlign(Style):
    """https://developer.mozilla.org/zh-CN/docs/Web/CSS/text-align"""

    def __init__(self, value="left") -> None:
        super().__init__()

        def fn():
            return {"text-align": value}

        self._dict_fn.append(fn)


class background(Style):
    """https://developer.mozilla.org/zh-CN/docs/Web/CSS/background"""

    def __init__(self, value="inherit") -> None:
        super().__init__()

        def fn():
            return {"background": value}

        self._dict_fn.append(fn)


class color(Style):
    """https://developer.mozilla.org/zh-CN/docs/Web/CSS/color"""

    def __init__(self, value="inherit") -> None:
        super().__init__()

        def fn():
            return {"color": value}

        self._dict_fn.append(fn)


class fontWeight(Style):
    """https://developer.mozilla.org/zh-CN/docs/Web/CSS/font-weight"""

    def __init__(self, value="400") -> None:
        super().__init__()

        def fn():
            return {"font-weight": value}

        self._dict_fn.append(fn)


class fontSize(Style):
    """https://developer.mozilla.org/zh-CN/docs/Web/CSS/font-size"""

    def __init__(self, value="1rem") -> None:
        super().__init__()

        def fn():
            return {"font-size": value}

        self._dict_fn.append(fn)


class lineHeight(Style):
    """https://developer.mozilla.org/zh-CN/docs/Web/CSS/line-height"""

    def __init__(self, value="1.5rem") -> None:
        super().__init__()

        def fn():
            return {"line-height": value}

        self._dict_fn.append(fn)



from pybi.core.components.componentTag import ComponentTag
from pybi.core.components.staticComponent import (
    TextComponent,
    UploadComponent,
    IconComponent,
    SvgIconComponent,
    SpaceComponent,
)
from pybi.core.components.reactiveComponent import (
    ReactiveComponent,
    Slicer,
    Table,
    TextValue,
    Markdown,
    Input,
    NumberSlider,
)
from pybi.core.components.containerComponent import (
    ContainerComponent,
    BoxComponent,
    FlowBoxComponent,
    GridBoxComponent,
    TabsComponent,
    AffixComponent,
)

from pybi.core.components.mermaid import Mermaid



#!/usr/bin/env python
"""The setup script."""

from setuptools import setup, find_packages
import pybi as pbi
from pathlib import Path


def all_static_files():
    files = [str(p) for p in Path("pybi/static").glob("*.*")]
    return files


with open("README.md", encoding="utf8") as readme_file:
    readme = readme_file.read()

requirements = ["pandas", "simplejson", "jinja2"]

test_requirements = ["pytest>=3", "playwright", "pyecharts"]

setup(
    author="carson_jia",
    author_email="568166495@qq.com",
    python_requires=">=3.7",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="...",
    entry_points={
        # 'console_scripts': [
        #     'test_prj=test_prj.cli:main',
        # ],
    },
    install_requires=requirements,
    license="MIT license",
    # long_description=readme,
    include_package_data=True,
    keywords=["pybi", "vision", "BI", "report"],
    name="pybi-next",
    packages=find_packages(include=["pybi", "pybi.*"]),
    data_files=[
        (
            "template",
            [
                "pybi/template/index.html",
            ],
        ),
        (
            "static",
            all_static_files(),
        ),
    ],
    test_suite="__tests",
    tests_require=test_requirements,
    url="",
    version=pbi.__version__,
    zip_safe=False,
)



from __future__ import annotations
from typing import TYPE_CHECKING, List, Optional, Union, Dict
from pybi.core.components.component import Component


from pybi.utils.data_gen import Jsonable, get_global_id
from .componentTag import ComponentTag


class Mermaid(Component):
    def __init__(self, graph: str) -> None:
        super().__init__(ComponentTag.Mermaid)
        self.graph = graph



from typing import TYPE_CHECKING, List, Dict, Any
from itertools import chain


def _extract_path(path: str):
    """
    >>> _extract_path('series[0]')
    >>> ['series',0]

    >>> _extract_path('data')
    >>> ['data']
    """
    if path[-1] == "]":
        arr = path.split("[")
        yield arr[0]
        yield int(arr[1][:-1])
    else:
        yield path


def get_by_paths(paths: List[str], data: Dict):
    """
    >>> dict_data = {
        'series':[
            {
                'data':[1,2,3,4]
            }
        ]
    }
    >>> get_by_paths(['series[0]','data'],dict_data)
    >>> [1,2,3,4]
    """

    if len(paths) == 0:
        return data

    ex_paths = chain.from_iterable(_extract_path(p) for p in paths)

    target = data[next(ex_paths)]

    for path in ex_paths:
        target = target[path]

    return target


def set_by_paths(paths: List[str], data: Dict, value: Any):
    """
    >>> dict_data = {
        'series':[
            {
                'data':[1,2,3,4]
            }
        ]
    }
    >>> set_by_paths(['series[0]','data'],dict_data,'changed')
    >>> dict_data
    >>> {
        'series':[
            {
                'data':'changed'
            }
        ]
    }
    """

    if len(paths) == 0:
        return

    ex_paths = list(chain.from_iterable(_extract_path(p) for p in paths))

    if len(ex_paths) == 1:
        data[ex_paths[0]] = value
        return

    target = data[ex_paths[0]]

    for p in ex_paths[1:-1]:
        target = target[p]

    target[ex_paths[-1]] = value



from __future__ import annotations
import pybi.utils.sql as sqlUtils
from pybi.utils.data_gen import Jsonable
from typing import Dict, Generator, Optional, List

import re

m_sql_from_text_pat = re.compile(
    r"(?:info:\[\|sql:)(.+?)\|type:(.+?)\|js_map:(.*?)(?:\|])", re.I | re.DOTALL
)


class SqlWrapper(Jsonable):
    def __init__(self, sql: str) -> None:
        super().__init__()
        self._sql_info = SqlInfo(sql)

    def split_group(self, by: str, code: str):
        code = f"""
function temp(key, rows, index){{
    {code}
}}
const gp = utils.groupBy(rows,'{by}')
const res =  Object.entries(gp).map(([key, rows], index) => temp(key,rows,index))
return res
        """
        return self.js_map(code)

    def js_map(self, code: str):
        """
        code: js code
            js vars:
                `rows` is object : query result rows,each row is object
                `fields` is string array : query result field names
                `utils` utils functions : `{uniq,groupBy}`
        ---
        pbi.sql(f"select name,age,value from {data}")
            `rows` e.g `[{name:'x1',age:20,value:100},{name:'x2',age:30,value:200},...]`
            `fields` e.g `['name','age','value']`
        """
        self._sql_info.jsMap = code
        self._sql_info.set_udf_type()
        return self

    def toflatlist(self, with_header=False):
        """
        >>> query = pbi.sql('select name,value from {data}')
        >>> query.toflatlist()
        >>> [
                ['name a',100],
                ['name b',200],
                ...
            ]

        >>> query.toflatlist(True)
        >>> [
                ['name','value'],
                ['name a',100],
                ['name b',200],
                ...
            ]
        """

        cols_define = "const cols=[fields]" if with_header else "const cols=[]"

        self.js_map(
            f"""
{cols_define}
if (fields.length>1){{
    return [...cols,...rows.map(data=> fields.map(f=> data[f]))] 
}}else {{
    return [...cols,...rows.map(data=>data[fields[0]])]
}}"""
        )
        return self

    def __str__(self) -> str:
        return str(self._sql_info)

    def _to_json_dict(self):
        return self._sql_info


class SqlInfo(Jsonable):
    def __init__(
        self, sql: str, type: str = "infer", js_map: Optional[str] = None
    ) -> None:
        """
        type:
            'udf' : user defined function
            'infer' : automatically extrapolate from the results
        """
        super().__init__()
        self.sql = sql
        self.type = type
        self.jsMap = js_map

    def set_udf_type(self):
        self.type = "udf"
        return self

    def __str__(self) -> str:
        return f"info:[|sql:{self.sql}|type:{self.type}|js_map:{self.jsMap or ''}|]"

    @staticmethod
    def extract_sql_from_text(text: str):
        """
        >>> input = '总销售额:sql:[_ select sum(销售额) from data _]'
        >>> extract_sql_from_text(input)
        >>> ['总销售额:',Sql('select sum(销售额) from data')]
        """
        start_idx = 0

        for match in re.finditer(m_sql_from_text_pat, text):

            span = match.span()

            if span[0] > start_idx:
                # 前面有普通文本
                yield text[start_idx : span[0]]

            js_map = match.group(3)
            yield SqlInfo(
                match.group(1), match.group(2), None if len(js_map) == 0 else js_map
            )
            start_idx = span[1]

        end_idx = len(text) - 1

        if start_idx <= end_idx:
            yield text[start_idx : len(text)]

    @staticmethod
    def around_backticks(text: str):
        def add_backtick(match: re.Match):
            return f"`{match.group()}`"

        return m_sql_from_text_pat.sub(add_backtick, text)


# class SqlExtractor:
#     js_code_flag = "--x_x--0_0--"

#     @staticmethod
#     def extract_from_dict(
#         data: Dict,
#         out_datasets: List[tuple[str, SqlWrapper]],
#         out_jscodes: List[tuple[str, str]],
#     ):
#         stack = [(data, "")]

#         while len(stack) > 0:
#             target, path = stack.pop()

#             if isinstance(target, dict):
#                 inputs = ((value, f"{path}.{key}") for key, value in target.items())
#                 stack.extend(inputs)
#                 continue

#             if isinstance(target, list):
#                 inputs = ((value, f"{path}[{idx}]") for idx, value in enumerate(target))
#                 stack.extend(inputs)
#                 continue

#             if isinstance(target, SqlWrapper):
#                 out_datasets.append((path[1:], target))

#             if (
#                 isinstance(target, str)
#                 and target[: len(SqlExtractor.js_code_flag)]
#                 == SqlExtractor.js_code_flag
#                 and target[-len(SqlExtractor.js_code_flag) :]
#                 == SqlExtractor.js_code_flag
#             ):
#                 target = target[
#                     len(SqlExtractor.js_code_flag)  : -len(SqlExtractor.js_code_flag)
#                 ]
#                 out_jscodes.append((path[1:], target))



from pybi.utils.data_gen import Jsonable, get_global_id


from typing import Optional


class StyleTagInfo(Jsonable):
    def __init__(
        self, css: str, id: Optional[str] = None, media: Optional[str] = None
    ) -> None:
        self.css = css
        self._media = media
        self._id = id

    def _to_json_dict(self):
        data = super()._to_json_dict()

        if self._media:
            data["media"] = self._media

        if self._id:
            data["id"] = self._id
        return data
